# Copyright (c) 2001-2005 Twisted Matrix Laboratories.
# Copyright (c) 2005 James Bunton
# See LICENSE for details.


"""An epoll() based implementation of the twisted main loop.

To install the event loop (and you should do this before any connections,
listeners or connectors are added)::

    from twisted.internet import epollreactor
    epollreactor.install()

API Stability: stable

Maintainer: U{James Bunton<mailto:james@delx.cjb.net>}
"""

# System imports
import errno, sys

# Twisted imports
from twisted.epoll import epoll
from twisted.python import log, threadable, failure
from twisted.internet import main, posixbase, error

# globals
reads = {}
writes = {}
masks = {}
selectables = {}
poller = epoll.ePoller()

POLL_READ = epoll.EPOLLIN | epoll.EPOLLPRI | epoll.EPOLLET
POLL_WRITE = epoll.EPOLLOUT | epoll.EPOLLET
POLL_DISCONNECTED = (epoll.EPOLLHUP | epoll.EPOLLERR)


class ePollReactor(posixbase.PosixReactorBase):
    """A reactor that uses epoll(2)."""

    def _retrieveFd(self, selectable):
        try:
            # the easy way
            fd = selectable.fileno()
            # make sure the fd is actually real.  In some situations we can get
            # -1 here.
            mdict[fd]
        except:
            # the hard way: necessary because fileno() may disappear at any
            # moment, thanks to python's underlying sockets impl
            for fd, fdes in selectables.items():
                if selectable is fdes:
                    break
            else:
                # Hmm, maybe not the right course of action?  This method can't
                # fail, because it happens inside error detection...
                return
        return fd

    def addReader(self, reader):
        """Add a FileDescriptor for notification of data available to read.
        """
        fd = reader.fileno()
        if not reads.has_key(fd):
            selectables[fd] = reader
            reads[fd] =  1
            mask = POLL_READ
            if writes.has_key(fd):
                mask |= POLL_WRITE
                poller.modregister(fd, mask)
            else:
                poller.register(fd, mask)

    def addWriter(self, writer, writes=writes, selectables=selectables):
        """Add a FileDescriptor for notification of data available to write.
        """
        fd = writer.fileno()
        if not writes.has_key(fd):
            selectables[fd] = writer
            writes[fd] =  1
            mask = POLL_WRITE
            if reads.has_key(fd):
                mask |= POLL_READ
                poller.modregister(fd, mask)
            else:
                poller.register(fd, mask)

    def removeReader(self, reader, reads=reads):
        """Remove a Selectable for notification of data available to read.
        """
        fd = self._retrieveFd(reader)
        if not fd: return

        if reads.has_key(fd): del reads[fd]
        if writes.has_key(fd):
            poller.modregister(fd, POLL_WRITE)
        else:
            if selectables.has_key(fd): del selectables[fd]
            poller.unregister(fd)

    def removeWriter(self, writer, writes=writes):
        """Remove a Selectable for notification of data available to write.
        """
        fd = self._retrieveFd(writer)
        if not fd: return

        if writes.has_key(fd): del writes[fd]
        if reads.has_key(fd):
            poller.modregister(fd, POLL_READ)
        else:
            if selectables.has_key(fd): del selectables[fd]
            poller.unregister(fd)

    def removeAll(self, reads=reads, writes=writes, selectables=selectables):
        """Remove all selectables, and return a list of them."""
        if self.waker is not None:
            self.removeReader(self.waker)
        result = selectables.values()
        fds = selectables.keys()
        reads.clear()
        writes.clear()
        selectables.clear()
        for fd in fds:
            poller.unregister(fd)
            
        if self.waker is not None:
            self.addReader(self.waker)
        return result

    def doPoll(self, timeout,
               reads=reads,
               writes=writes,
               selectables=selectables,
               log=log):
        """Poll the poller for new events."""
        if timeout is not None:
            timeout = int(timeout * 1000) # convert seconds to milliseconds

        try:
            l = poller.poll(timeout)
        except epoll.ePollError, e:
            if e.value == errno.EINTR:
                return
            else:
                raise
        _drdw = self._doReadOrWrite
        for fd, event in l:
            try:
                selectable = selectables[fd]
            except KeyError:
                # Handles the infrequent case where one selectable's
                # handler disconnects another.
                continue
            log.callWithLogger(selectable, _drdw, selectable, fd, event, log)

    doIteration = doPoll

    def _doReadOrWrite(self, selectable, fd, event, log, 
        faildict={
            error.ConnectionDone: failure.Failure(error.ConnectionDone()),
            error.ConnectionLost: failure.Failure(error.ConnectionLost())
        }):
        why = None
        inRead = False
        if event & POLL_DISCONNECTED and not (event & (epoll.EPOLLIN | epoll.EPOLLPRI)):
            why = main.CONNECTION_LOST
        else:
            try:
                if (event & (epoll.EPOLLIN | epoll.EPOLLPRI)):
                    why = selectable.doRead()
                    inRead = True
                if not why and event & epoll.EPOLLOUT:
                    why = selectable.doWrite()
                    inRead = False
                if not selectable.fileno() == fd:
                    why = error.ConnectionFdescWentAway('Filedescriptor went away')
                    inRead = False
            except:
                log.deferr()
                why = sys.exc_info()[1]
        if why:
            self._disconnectSelectable(selectable, why, inRead)


def install():
    """Install the epoll() reactor."""
    p = ePollReactor()
    import main
    main.installReactor(p)


__all__ = ["ePollReactor", "install"]
